package j2gether.model;

import java.io.*;
import java.util.Iterator;
import java.util.LinkedList;
import j2gether.view.listeners.ScoreListener;

/**
 * <p>Title: J2Gether</p>
 *
 * <p>Description: Primeiro trabalho prtico P4</p>
 *
 * <p>Copyright: Copyright (c) 2005</p>
 *
 * <p>Company: 4thFloor</p>
 *
 * @author Srgio Lopes, Ruben Pedro
 * @version 1.0
 */
public class Score implements Serializable {

  /**
   * Conjunto de nomes dos jogadores.
   */
  private String[] names;

  /**
   * Pontuao obtida pelos jogadores.
   */
  private double[] scores;

  /**
   * Atributo interno que indica em que posiao ser introduzido o novo par nome de jogador e pontuao.
   */
  private transient int position = -1;

  /**
   * Nome do ficheiro de pontuaes a gravar.
   */
  private transient String baseName;

  /**
   * Listeners de eventos para o Score.
   */
  private transient LinkedList listeners;

  /**
   * Constante com o nmero de erro para uma posio invlida
   */
  public static final int POS_ERR = -1;

  /**
   * Nome da directoria base onde sero guardados os ficheiros de pontuaes
   */
  private static String BASEDIR = ".J2Gether";

  /**
   * Nome da sub-directoria base onde sero guardados os ficheiros de pontuaes
   */
  private static String BASESUBDIR = "Score";

  /**
   * Constri uma tabela de pontuaes com base no nmero mximo de pontuaes e no modo
   *
   * @param numScores nmero mximo de pontuaes a guardar
   * @param mode modo no qual se baseia esta tabela de pontuaes
   * @see j2gether.model.ScoreMode
   */
  public Score(int numScores, ScoreMode mode) {
      names = new String[numScores];
      scores = new double[numScores];
      baseName = Integer.toString(mode.getNumLinhasTabuleiro() *
                                  mode.getNumColunasTabuleiro() *
                                  mode.getNumCores());
      ObjectInputStream inStream = null;

      try {
          File scoreFile = new File(System.getProperty("user.home") +
                                    File.separator + BASEDIR +
                                    File.separator + BASESUBDIR +
                                    File.separator +
                                    baseName + ".dat");
          if (scoreFile.exists()) {
              inStream = new ObjectInputStream(new FileInputStream(scoreFile));
              Score sc = (Score) inStream.readObject();
              if (sc != null) {
                  names = sc.names;
                  scores = sc.scores;
              }
          }
      } catch (ClassNotFoundException ex3) {
          System.out.println(ex3);
      } catch (IOException ex1) {
          System.out.println(ex1);
      } finally {
          if (inStream != null) {
              try {
                  inStream.close();
              } catch (IOException ex2) {
                  System.out.println(ex2);
              }
          }
      }
  }

  /**
   * Devolve a pontuao indicada por index.
   *
   * @param index ndice para o qual se pretende a pontuao
   * @return pontuao no ndice indicado
   */
  public double getScoreValueAt(int index) {
    return scores[index];
  }

  /**
   * Devolve o nome do jogador indicado por index.
   *
   * @param index ndice para o qual se pretende o nome do jogador
   * @return nome do jogador no ndice indicado
   */
  public String getScoreNameAt(int index) {
    return names[index];
  }

  /**
   * Obtm o ndice no qual sero introduzidos o nome do jogador e a sua pontuao
   *
   * @param index ndice para o qual se pretende a pontuao
   * @return posiso para a nova pontuao ou POS_ERR no caso de no ser possvel introduzir a nova pontuao
   */
  private int findScorePosition(double score) {
    for (int z = 0; z < names.length; z++) {
      if (scores[z] <= score) {
        return z;
      }
    }
    return POS_ERR;
  }

  /**
   * Insere um nome e uma pontuao no ndice indicado
   *
   * @param index ndice no qual se vo introduzir o nome e a pontuao
   * @param name nome a introduzir
   * @param score pontuao a introduzir
   */
  private void insertScore(int index, String name, double score) {
    if (index >= 0 && index < names.length) {
      names[index] = name;
      scores[index] = score;
    }
  }

  /**
   * Verifica se a pontuao indicada tem lugar na tabela de pontuaes.
   *
   * @param index ndice no qual se vo introduzir o nome e a pontuao
   * @return boolean indicando se  ou no possvel introduzir a nova pontuao
   */
  public boolean isInsertableScore(double score) {
    return (position = findScorePosition(score)) != -1;
  }

  /**
   * Permite introduzir um nome e uma pontuao nova. A posio onde os novos valores so introduzidos  indicada internamente.
   *  obrigatrio uma chamada ao mtodo isInsertableScore antes da utilizao deste mtodo.
   *
   * @param index ndice no qual se vo introduzir o nome e a pontuao
   * @return boolean indicando se  ou no possvel introduzir a nova pontuao
   * @see j2gether.model.Score#isInsertableScore(double)*/
  public void updateScores(String name, double score) {
    if (names[position] != null) {
      allocateSpace(position);
    }
    insertScore(position, name, score);
  }

  /**
   * Caso seja necessrio introduzir uma nova pontuao onde j se encontrem valores este mtodo tem por objectivo
   * alocar espao deslocando os dados para baixo
   *
   * @param stopAt ndice no qual o mtodo termina de deslocar dados
   */
  private void allocateSpace(int stopAt) {
    for (int z = names.length - 1; z >= stopAt; z--) {
      if (z != 0) {
        names[z] = names[z - 1];
        scores[z] = scores[z - 1];
      }
    }
  }

  /**
   * Devolve o nmero de espaos para pontuaes
   *
   * @return nmero total de pontuaes possveis de serem introduzidas
   */
  public int getSize() {
    return names.length;
  }

  /**
   * Redefinio do mtodo toString da classe Object
   */
  public String toString() {
    StringBuffer strbuff = new StringBuffer();
    for (int z = 0; z < names.length; z++) {
      strbuff.append(names[z] + " ");
      strbuff.append(scores[z]);
      strbuff.append("\n");
    }
    return strbuff.toString();
  }

  /**
   * Grava a tabela pontuaes corrente. Cria a directoria score, se esta no existir, onde grava o ficheiro
   *
   * @return boolean indicando se foi possvel gravar a tabela de pontuaes
   */
  public boolean saveScore() {
      ObjectOutputStream outStream = null;
      try {
          File scoreDir = new File(System.getProperty("user.home") + File.separator + BASEDIR + File.separator);
          if(!scoreDir.exists())
              scoreDir.mkdir();
          scoreDir = new File(System.getProperty("user.home") + File.separator + BASEDIR + File.separator + BASESUBDIR + File.separator);
          if (!scoreDir.exists())
              scoreDir.mkdir();
          File scoreFile = new File(System.getProperty("user.home") +
                                    File.separator + BASEDIR + File.separator + BASESUBDIR + File.separator +
                                    baseName + ".dat");
          if (!scoreFile.exists())
              scoreFile.createNewFile();
          outStream = new ObjectOutputStream(new FileOutputStream(scoreFile));
          outStream.writeObject(this);
          return true;
      } catch (IOException ex1) {
          System.out.println(ex1);
      } finally {
          if (outStream != null) {
              try {
                  outStream.close();
              } catch (IOException ex2) {
                  System.out.println(ex2);
              }
          }
      }
      return false;
  }

  /**
   * Permite eliminar todos os nomes e pontuaes antes guardadas
   */
  public void reset() {
    names = new String[names.length];
    scores = new double[scores.length];
  }

  /**
   * Adiciona um ScoreListener  lista de listeners
   *
   * @param l listener a ser adicionado
   * @see j2gether.view.listeners.ScoreListener
   */
  public void addScoreListener(ScoreListener l) {
    if (listeners == null) {
      listeners = new LinkedList();
    }
    listeners.add(l);
  }

  /**
   * Elimina um listener da lista
   *
   * @param l listener a ser removido
   * @see j2gether.view.listeners.ScoreListener
   */
  public void removeScoreListener(ScoreListener l) {
    if (listeners != null) {
      listeners.remove(l);
    }
  }

  /**
   * Mtodo que indica a todos os listeners registados que houve alteraes na tabela de pontuaes
   */
  public void fireScoreChanged() {
    Iterator it = listeners.iterator();
    while (it.hasNext()) {
      ((ScoreListener) it.next()).scoreChanged();
    }
  }
}
